from typing import Any
from typing import Literal

import numpy as np
from numpy.typing import NBitBase
from numpy.typing import NDArray

from mitosis._typing import ExpResults


def fit_and_score(
    data: NDArray[np.floating[NBitBase]], metric: Literal["len"] | Literal["zero"]
) -> ExpResults:
    if metric == "len":
        return {"main": len(data)}
    elif metric == "zero":
        return {"main": 0}


def bad_runnable(*args: Any, **kwargs: Any) -> int:
    return 1  # not a dict with key "main"



import pickle
import re
import sys
from pathlib import Path
from types import ModuleType
from typing import Callable

import nbclient.exceptions
import pytest

import mitosis
from mitosis import _disk
from mitosis import cleanstr
from mitosis import unpack
from mitosis._typing import ExpStep
from mitosis._typing import Parameter
from mitosis.tests import mock_paper
from mitosis.tests import mock_part1
from mitosis.tests import mock_part2


def mock_callable():
    pass


class MockContainer:
    a: list[Callable]

    def __init__(self, a):
        self.a = a


def test_custom_container():
    fakeobj = MockContainer([mock_callable])
    assert "0x" not in cleanstr(fakeobj)


def test_str_int_str():
    assert "'" in cleanstr("1") or '"' in cleanstr("1")
    assert "'" not in cleanstr(1) and '"' not in cleanstr(1)


def test_reproduceable_dict():
    mydict = mitosis.StrictlyReproduceableDict(**{"1": print})
    assert str(mydict) == r"{'1': <builtin_function_or_method builtins.print>}"


def test_reproduceable_list():
    mylist = mitosis.StrictlyReproduceableList([1, print])
    assert str(mylist) == r"[1, <builtin_function_or_method builtins.print>]"


def test_unreproduceable_list():
    # test function in a local closure
    with pytest.raises(ValueError):
        str(mitosis.StrictlyReproduceableList([1, lambda x: 1]))


def test_reproduceable_list_of_strs():
    mylist = mitosis.StrictlyReproduceableList(["a"])
    assert str(mylist) == r"['a']"


def test_reproduceable_dict_of_strs():
    mylist = mitosis.StrictlyReproduceableDict({"a": "b"})
    assert str(mylist) == r"{'a': 'b'}"


def test_nested_reproduceable_classes():
    mylist = mitosis.StrictlyReproduceableList([print])
    mylist = mitosis.StrictlyReproduceableList([mylist])
    mydict = mitosis.StrictlyReproduceableDict(a=mylist)
    mydict = mitosis.StrictlyReproduceableDict(b=mydict)
    result = str(mydict)
    assert result == r"{'b': {'a': [[<builtin_function_or_method builtins.print>]]}}"


def mock_global_f():
    pass


mock_global_f.__module__ = "__main__"


def test_unreproduceable_dict():
    # test function in a local closure
    def mock_local_f():
        pass

    with pytest.raises(ImportError):
        str(mitosis.StrictlyReproduceableDict(**{"1": mock_local_f}))

    # test function defined in __main__
    mock_global_f.__module__ = "__main__"
    with pytest.raises(ImportError):
        str(mitosis.StrictlyReproduceableDict(**{"1": mock_global_f}))
    mock_global_f.__module__ = __name__

    # test unimportable module
    newmod = ModuleType("_mockmod")
    setattr(newmod, "mock_global_f", mock_global_f)
    mock_global_f.__module__ = newmod.__name__
    with pytest.raises(ImportError):
        str(mitosis.StrictlyReproduceableDict(**{"1": mock_global_f}))
    mock_global_f.__module__ = __name__

    # test module missing name
    newmod = ModuleType("_mockmod")
    mock_global_f.__module__ = newmod.__name__
    sys.modules["_mockmod"] = newmod
    with pytest.raises(ImportError):
        str(mitosis.StrictlyReproduceableDict(**{"1": mock_global_f}))
    mock_global_f.__module__ = __name__
    sys.modules.pop("_mockmod")

    # test lambda function
    with pytest.raises(ValueError):
        str(mitosis.StrictlyReproduceableDict(**{"1": lambda x: 1}))


@pytest.fixture
def fake_eval_param():
    return mitosis.Parameter("1", "seed", 1, evaluate=True)


@pytest.fixture
def fake_lookup_param():
    return mitosis.Parameter("test", "foo", 2, evaluate=False)


@pytest.fixture()
def mock_steps():
    return [
        # fmt: off
        ExpStep(
            "foo",
            mock_part1.Klass.gen_data, "mitosis.tests.mock_part1:Klass.gen_data",
            mock_paper.data_config, "mitosis.tests.mock_paper:data_config",
            None,
            [
                Parameter("test", "length", 5, evaluate=False),
                Parameter("True", "extra", True, evaluate=True),
            ],
            []
        ),
        ExpStep(
            "bar",
            mock_part2.fit_and_score, "mitosis.tests.mock_part2:fit_and_score",
            mock_paper.meth_config, "mitosis.tests.mock_paper:meth_config",
            None,
            [
                Parameter("test", "metric", "len", evaluate=False),
            ],
            []
        )
        # fmt: on
    ]


def test_mock_experiment(mock_steps, tmp_path):
    exp_key = mitosis.run(
        mock_steps,
        debug=True,
        trials_folder=tmp_path,
    )
    data = mitosis.load_trial_data(exp_key, trials_folder=tmp_path)
    assert len(data[0]["data"]) == 5
    params = mitosis._load_trial_params(exp_key, step=0, trials_folder=tmp_path)
    assert params == {"length": 5, "extra": True}
    metadata = mitosis._disk.locate_trial_folder(exp_key, trials_folder=tmp_path)
    assert (metadata / "experiment").resolve().exists()


@pytest.fixture
def nothing_step():
    # fmt: off
    return ExpStep(
        "nothing",
        mock_part1.do_nothing, "mitosis.tests.mock_part1:do_nothing",
        mock_paper.lookup_default, "mitosis.tests.mock_paper:lookup_default",
        None,
        [],
        []
    )
    # fmt: on


@pytest.mark.clean
def test_variant_redefinition_disallowed(nothing_step, tmp_path):
    # GH 56
    chg_param1 = Parameter("foo_a", "foo", "a", evaluate=False)
    chg_param2 = Parameter("foo_a", "foo", "b", evaluate=False)
    nothing_step.args.append(chg_param1)
    mitosis.run([nothing_step], trials_folder=tmp_path)
    nothing_step.args.pop(0)
    nothing_step.args.append(chg_param2)
    with pytest.raises(RuntimeError, match="stored with different values"):
        mitosis.run([nothing_step], trials_folder=tmp_path)


def test_load_results_order(tmp_path):
    exp_key = "test_results"
    (tmp_path / exp_key).mkdir()
    filename1 = "results1.dill"
    filename2 = "results0.dill"
    with open(tmp_path / exp_key / filename1, "wb") as f1:
        pickle.dump("second", f1)
    with open(tmp_path / exp_key / filename2, "wb") as f2:
        pickle.dump("first", f2)
    result = mitosis.load_trial_data(exp_key, trials_folder=tmp_path)
    expected = ["first", "second"]
    assert result == expected


def test_mod_metadata_debug(mock_steps, tmp_path):
    hexstr = mitosis.run(
        mock_steps,
        debug=True,
        trials_folder=tmp_path,
    )
    trial_folder = _disk.locate_trial_folder(hexstr, trials_folder=tmp_path)
    with open(trial_folder / "experiment.log", "r") as f:
        log_str = "\n".join(f.readlines())
    assert len(re.findall("This is run every time", log_str)) == 1
    assert len(re.findall("This is run in debug mode only", log_str)) == 1
    with open(trial_folder / "config.txt") as f:
        config_lines = f.readlines()
    config_params = [eval(line) for line in config_lines]
    passed_params = [{p.arg_name: p.vals for p in step.args} for step in mock_steps]
    assert config_params == passed_params


@pytest.mark.clean
def test_mod_metadata(mock_steps, tmp_path):
    hexstr = mitosis.run(
        mock_steps,
        debug=False,
        trials_folder=tmp_path,
    )
    trial_folder = _disk.locate_trial_folder(hexstr, trials_folder=tmp_path)
    with open(trial_folder / "experiment.log", "r") as f:
        log_str = "\n".join(f.readlines())
    assert len(re.findall("This is run every time", log_str)) == 1
    assert len(re.findall("This is run in debug mode only", log_str)) == 0
    with open(trial_folder / "config.txt") as f:
        config_lines = f.readlines()
    config_params = [eval(line) for line in config_lines]
    passed_params = [{p.arg_name: p.vals for p in step.args} for step in mock_steps]
    assert config_params == passed_params


def test_malfored_return_experiment(mock_steps, tmp_path):
    bad_steps = [
        mock_steps[0],
        ExpStep(
            mock_steps[1].name,
            mock_part2.bad_runnable,  # type: ignore
            "mitosis.tests.mock_part2:bad_runnable",
            mock_steps[1].lookup,
            mock_steps[1].lookup_ref,
            mock_steps[1].group,
            mock_steps[1].args,
            mock_steps[1].untracked_args,
        ),
    ]
    with pytest.raises(nbclient.exceptions.CellExecutionError):
        mitosis.run(
            bad_steps,
            debug=True,
            trials_folder=tmp_path,
        )


def test_load_toml():
    parent = Path(__file__).resolve().parent
    tomlfile = parent / "test_pyproject.toml"
    result = _disk.load_mitosis_steps(tomlfile)["nothing"]
    expected = (
        "mitosis.tests.mock_part1:do_nothing",
        "mitosis.tests.mock_paper:lookup_default",
    )
    assert result == expected


def test_load_bad_toml():
    parent = Path(__file__).resolve().parent
    tomlfile = parent / "pyproject_missing.toml"
    with pytest.raises(RuntimeError, match="does not have a tool"):
        _disk.load_mitosis_steps(tomlfile)
    tomlfile = parent / "pyproject_malformed.toml"
    with pytest.raises(RuntimeError, match="table is malformed"):
        _disk.load_mitosis_steps(tomlfile)


def test_unpack():
    from importlib.metadata import version

    obj_ref = "importlib.metadata:version"
    result = unpack(obj_ref)
    assert result is version



from functools import lru_cache
from pathlib import Path
from typing import Any
from typing import Optional

import git
import toml


@lru_cache
def load_mitosis_steps(
    proj_file: Path | str = "pyproject.toml",
) -> dict[str, tuple[str, str]]:
    """Load experiment steps saved in pyproject.toml (or other config)

    Args:
        pyproj_file:
            Local or absolute path to pyproject file.  default is pyproject.toml
            if local path or default, use current git repo's top level directory
    Raises:
        Runtime error if cannot load steps, or if they are badly formed.
    """
    proj_file = _choose_toml(proj_file)
    config = _load_config(proj_file)
    try:
        result = config["steps"]
    except KeyError:
        raise RuntimeError(
            f"{proj_file.absolute()} does not have a tool.mitosis.steps table"
        )
    if any(not isinstance(vals, list) or len(vals) != 2 for vals in result.values()):
        raise RuntimeError("tool.mitosis.steps table is malformed")
    return {k: tuple(v)[:2] for k, v in result.items()}


@lru_cache
def get_repo() -> git.Repo:
    repo = git.Repo(Path.cwd(), search_parent_directories=True)
    return repo


@lru_cache
def _choose_toml(filename: Path | str | None) -> Path:
    """Identify the absolute location of the project file"""
    repo = get_repo()
    directory = Path(repo.working_dir).absolute()
    if filename is None:
        return directory / "pyproject.toml"
    elif not Path(filename).is_absolute():
        return directory / filename
    return Path(filename)


def _load_config(toml_pth: Path) -> dict[str, Any]:
    """Load the mitosis section of the config dictionary

    Raises:
        KeyError if no tool.mitosis table exists
    """
    if not toml_pth.is_absolute():
        raise ValueError("Resolve path prior to final loading step")
    with open(toml_pth, "rt", encoding="utf-8") as f:
        config = toml.load(f)
    return config["tool"]["mitosis"]


def locate_trial_folder(
    hexstr: Optional[str] = None,
    *,
    trials_folder: Optional[Path | str] = None,
    proj_file: str = "pyproject.toml",
) -> Path:
    """Identify where trials are saved.

    Args:
        hexstr: the hexstring of a trial.  Default (None) returns the
            directory of all trials.  If not none, returns the metadata
            folder for that trial
        trials_folder: relative or absolute path of trials folder.  Default
            (None) looks for a location in the project config file, or if
            missing, uses <repository root> / 'trials'.
        proj_file: path to the toml config for this project.
    """
    if trials_folder is None:
        try:
            config = _load_config(_choose_toml(proj_file))
            trials_folder = Path(config["trials-folder"])
        except KeyError:
            trials_folder = Path("trials")
    else:
        trials_folder = Path(trials_folder).resolve()
    if not trials_folder.is_absolute():
        trials_folder = Path(get_repo().working_dir).absolute() / trials_folder
    if not hexstr:
        return trials_folder

    matches = trials_folder.glob(f"*{hexstr}")
    try:
        first = next(matches)
    except StopIteration:
        raise FileNotFoundError(f"Could not find a trial that matched {hexstr}")
    try:
        next(matches)
    except StopIteration:
        return first
    raise RuntimeError(f"Two or more matches found for {hexstr}")



from logging import getLogger

import numpy as np

from mitosis._typing import ExpResults


class Klass:
    @staticmethod
    def gen_data(length: int, extra: bool = False) -> ExpResults:
        getLogger(__name__).info("This is run every time")
        getLogger(__name__).debug("This is run in debug mode only")

        return {
            "data": np.ones(length, dtype=np.float_),
            "extra": extra,
            "main": None,
        }  # type: ignore


def do_nothing(*args, **kwargs) -> ExpResults:
    """An experiment step that accepts anything and produces nothing"""
    return {"main": None}



from collections import defaultdict

data_config = {"length": {"test": 5}}

meth_config = {"metric": {"test": "len"}}

# lookup any parameter, any variant: always none
lookup_default: dict[str, dict[str, None]] = defaultdict(
    lambda: defaultdict(lambda: None)
)



from collections.abc import Mapping
from dataclasses import dataclass
from dataclasses import field
from typing import Callable
from typing import NamedTuple
from typing import ParamSpec
from typing import TypedDict


class ExpResults(TypedDict):
    main: object


P = ParamSpec("P")
ExpRun = Callable[P, ExpResults]


@dataclass
class Parameter:
    """An experimental parameter

    Arguments:
        var_name: short name for the variant (particular values) across use cases
        arg_name: name of arg known to experiment
        vals: value of the parameter
        eval: whether variant name should be evaluated or looked up.
    """

    var_name: str
    arg_name: str
    vals: object
    # > 3.10 only: https://stackoverflow.com/a/49911616/534674
    evaluate: bool = field(default=False, kw_only=True)


class ExpStep(NamedTuple):
    name: str
    action: ExpRun
    action_ref: str
    lookup: Mapping[str, Mapping[str, object]]
    lookup_ref: str
    group: str | None
    args: list[Parameter]
    untracked_args: list[str]



from pathlib import Path
from typing import Optional

from sqlalchemy import Column
from sqlalchemy import create_engine
from sqlalchemy import Engine
from sqlalchemy import Float
from sqlalchemy import Insert
from sqlalchemy import insert
from sqlalchemy import inspection
from sqlalchemy import Integer
from sqlalchemy import MetaData
from sqlalchemy import String
from sqlalchemy import Table
from sqlalchemy import Update
from sqlalchemy import update

from mitosis._typing import Parameter


def trials_columns() -> list[Column]:
    return [
        Column("variant", String, primary_key=True),
        Column("iteration", Integer, primary_key=True),
        Column("commit", String, nullable=False),
        Column("cpu_time", Float),
        Column("results", String),
        Column("filename", String),
    ]


def variant_types() -> list[Column]:
    return [
        Column("name", String, primary_key=True),
        Column("params", String, unique=False),
    ]


def create_trials_db_eng(filename: Path | str, table_name: str) -> tuple[Engine, Table]:
    if Path(filename).is_absolute():
        db = Path(filename)
    else:
        db = Path(__file__).resolve().parent / filename

    md = MetaData()
    table = Table(table_name, md, *trials_columns())
    url = "sqlite:///" + str(db)
    eng = create_engine(url)
    with eng.begin() as conn:
        if not inspection.inspect(conn).has_table(table_name):
            md.create_all(conn)
    return eng, table


def record_start_in_db(
    tb: Table, eng: Engine, variant: str, iteration: int, commit: str
) -> Insert:
    stmt = insert(tb)
    stmt = stmt.values({"variant": variant, "iteration": iteration, "commit": commit})
    with eng.begin() as conn:
        conn.execute(stmt)
    return stmt


def record_finish_in_db(
    tb: Table,
    eng: Engine,
    variant: str,
    iteration: int,
    metric: Optional[str],
    filename: str,
    total_time: float,
) -> Update:
    stmt = update(tb)
    # primary keys... type ignore is only needed in mypy, not pyright
    stmt = stmt.where(tb.columns.get("variant") == variant)  # type: ignore
    stmt = stmt.where(tb.columns.get("iteration") == iteration)  # type: ignore
    stmt = stmt.values(
        {
            tb.columns.get("cpu_time"): total_time,
            tb.columns.get("results"): metric,
            tb.columns.get("filename"): filename,
        }
    )
    with eng.begin() as conn:
        conn.execute(stmt)
    return stmt


def _init_variant_table(trial_db: Path, step: str, param: Parameter) -> Table:
    eng = create_engine("sqlite:///" + str(trial_db))
    md = MetaData()
    var_table = Table(f"{step}_variant_{param.arg_name}", md, *variant_types())
    inspector = inspection.inspect(eng)
    if not inspector.has_table(f"{step}_variant_{param.arg_name}"):
        md.create_all(eng)
    return var_table
